CONSTANT_DATASET: str = "preql_internal_constant_dataset"
ALL_ROWS_CONCEPT = "all_rows"
INTERNAL_NAMESPACE = "__preql_internal"
PERSISTED_CONCEPT_PREFIX = "__pre_persist"
UNNEST_NAME = "_unnest_alias"
